{{
        Continuous Time Stamping on the propeller

        Jonathan "lonesock" Dummer

        Note: I measure 2x periods at a time, because
        I wanted rates above 2MHz, and could not sample
        faster than at about 1.5 MHz when capturing
        single periods.

        Using a Least Squares Error fit:
                timestamp(i) = Period * i
        so, LSE gives a matrix like:

        [ sum( 1 )  sum( i ) ]   { offset }   { sum(  ts  ) }
        [                    ] * {        } = {             }
        [ sum( i )  sum(i*i) ]   { period }   { sum( ts*i ) }

        I use variable names X and Y instead of i and timestamp.
        Cramer's Rule for solving for the period is:
                period = (sum1*sumYX - sumX*sumY) / (sum1*sumXX - sumX*sumX)

        Bean pointed out that:
                sumX  = sum1 * (sum1 + 1) / 2
                sumXX = sum1 * (sum1 + 1) * (2 * sum1 + 1) / 6

        Using those substitutions, I get:
                period = (12*sumYX - 6*(sum1+1)*sumY) / (sum1*(sum1+1)*(sum1-1))

        Note that internally I actualy do timestamps every 2 periods
        to increase the max sampleable frequency.  This means I would
        have been solving for 2*Period, so instead I simply modified
        the numerator of the above equation to use a 6 and 3 instead
        of 12 and 6.

        Finally, to get the actual frequency, the user would
        divide clkfreq by the period.  Instead of trying that,
        I simply generate the inverse of the period as a 32-bit
        fraction.  Thatway, recovering the frequency in Spin is
        as easy as:
                freq := clkfreq ** inv_period_frac
}}

CON
  bufsize = 1024  ' we may not use all samples...during the calcs, if the integers are about to overflow it stops computiing
'DAT
'  buf   long  0[ bufsize + 1 ]
VAR
  byte  cog
  long  buf[ bufsize + 1 ]

PUB start( pin, timeout_clocks, ptr_to_sum_1_x_xx_y_yx )
  pin_mask := |<pin
  timeout := timeout_clocks
  buf_ptr_sub4 := @buf - 4                    
  return cog := cognew( @CTS_PASM_entry, ptr_to_sum_1_x_xx_y_yx ) + 1

PUB stop
  if cog
    cogstop( cog~ - 1 )


DAT
ORG 0
CTS_PASM_entry          ' Set up the variables
                        mov     ptr, buf_ptr_sub4
                        mov     i, buf_size
                        mov     tmp, #0

                        ' mark the time of the first edge
                        waitpne pin_mask, pin_mask
                        waitpeq pin_mask, pin_mask
                        mov     timein, cnt
                        
                        ' wait 1 more edge
                        waitpne pin_mask, pin_mask
                        waitpeq pin_mask, pin_mask


get_samples             ' use this time to get some work done
                        add     ptr, #4
                        cmps    tmp, timeout wc         ' this test is one loop (2 periods) out of date
                        neg     tmp, timein             ' pre-adjust to the start time
                        
                        ' wait till low, then wait for a high edge
                        waitpne pin_mask, pin_mask                
                        waitpeq pin_mask, pin_mask
                        
                        ' grab the time, reference to the start time, store it
                        add     tmp, cnt                                                
                        wrlong  tmp, ptr

                        ' wait till low, then wait for a high edge
                        waitpne pin_mask, pin_mask                
                        waitpeq pin_mask, pin_mask                   

                        ' if we still have time and space, go back for another sample                                          
              if_c      djnz    i, #get_samples
               
                        ' done gathering, mark a final 0
                        add     ptr, #4 
                        wrlong  dirb, ptr


Process_values
                        ' Process!
                        mov     sum1,   #0
                        mov     sumY,   #0
                        mov     sumYX,  #0
                        mov     sumYXhi,#0
                        mov     ptr, buf_ptr_sub4
                        ' get each one back again
proc_loop               add     ptr, #4
                        rdlong  tmp, ptr wz
              if_z      jmp     #done_proc
                        ' do the LSE accumulations
                        add     sum1, #1
                        add     sumY, tmp
                        ' this is the biggest one, so use 64 bits
                        mov     v1, sum1
                        mov     v2, tmp
                        call    #multu64
                        add     sumYX, vRes wc
                        addx    sumYXhi, vResHi

                        ' more to do?
                        jmp     #proc_loop

done_proc               ' now compute the inverse period ratio
                        ' Period Numerator first

                        ' 2 * sumYX (all 64 bits)
                        shl     sumYX, #1 wc
                        rcl     sumYXhi, #1
                        ' (sum1 + 1) * sumY
                        mov     v1, sum1
                        add     v1, #1
                        mov     v2, sumY
                        call    #multu64
                        ' subtract
                        sub     sumYX, vRes wc
                        subx    sumYXhi, vResHi
                        ' multiply by 3
                        mov     tmp, sumYXhi
                        add     sumYXhi, tmp
                        add     sumYXhi, tmp
                        mov     tmp, sumYX
                        add     sumYX, tmp wc
                        addx    sumYXhi, #0
                        add     sumYX, tmp wc
                        addx    sumYXhi, #0

                        ' Denominator next:
                        ' (sum1 - 1) * (sum1 + 1)
                        mov     v1, sum1
                        sub     v1, #1
                        mov     v2, sum1
                        add     v2, #1
                        call    #multu64
                        ' *= sum1
                        mov     v1, sum1
                        mov     v2, vRes
                        call    #multu64
                        ' Period Denominator is now in vRes and vResHi


                        ' Period now = sumYX / vRes
                        ' compute the _inverse_ ratio!
                        mov     i, #32
:get_ratio              ' shift my result left by 1
                        shl     tmp, #1
                        ' multiply vRes by 2
                        shl     vRes, #1 wc
                        rcl     vResHi, #1
                        ' compare it to sumYX
                        cmp     sumYX, vRes wc,wz
                        cmpx    sumYXhi, vResHi wc,wz
                        ' store that bit
              if_a      jmp     #:cont_ratio
                        ' do the subtraction
                        sub     vRes, sumYX wc
                        subx    vResHi, sumYXhi
                        ' mark that we did a subtraction
                        or      tmp, #1
                        ' do this all 32 times
:cont_ratio             djnz    i, #:get_ratio

                        ' was always rounding down...HACK
                        add     tmp, #2

                        ' Save it!
                        wrlong  tmp, par

                        ' just start over again
                        jmp     #CTS_PASM_entry


multu64sort             ' shuffle the values so v1 has fewer high 1 bits
                        mov     vRes, v1
                        max     v1, v2
                        min     v2, vRes
multu64                 ' already sorted (call here if you know v1 is smaller)
                        mov     vRes, #0
                        mov     vResHi, #0
                        mov     v2Hi, #0
                        ' begin the core multiplication loop
:loop                   shr     v1, #1 wc, wz           ' get the lowest bit of v1
              if_c      add     vResHi, v2Hi
              if_c      add     vRes, v2 wc             ' if it was a 1, add v2 to the accumulator
                        addx    vResHi, #0
                        shl     v2, #1 wc               ' shift v2 left by 1 (to prep for the next v1 bit)
                        rcl     v2Hi, #1
              if_nz     jmp     #:loop                  ' loop as long as v1 is not 0
multu64sort_ret
multu64_ret             ret
                
                        
pin_mask                long    0
buf_ptr_sub4            long    0
buf_size                long    bufsize
timeout                 long    0

' accumulators for the Least Squares Error calcs
sum1                    res     1
sumY                    res     1
sumYX                   res     1
sumYXhi                 res     1

' local vars
ptr                     res     1
i                       res     1
tmp                     res     1
timein                  res     1
v1                      res     1
v2                      res     1
vRes                    res     1
v2Hi                    res     1
vResHi                  res     1
vTmp                    res     1